/**  
 * Pikku esimerkki deadlockista.
 * 
 * 
 * Esimerkki luennolta vuoden 2015 vapun alla. Vuonna 2016 vasta
 * viikko vapun jälkeen, mutta eiköhän vappu muisteta vielä, niin
 * saadaan tästä vappusimulaattorista irti huvi hyödyn lisäksi.
 *
 *  Tehdään kaksi säiettä, jotka juovat simaa ja syövät munkkia.
 *  Molemmat resurssit ovat yhteisessä pöydässä, joten simasammio ja
 *  munkkilaari täytyy lukita, että yhdistetty vappunautinto onnistuu.
 *
 * Tässä esimerkissä vappu ei jumiudu, koska siman lukitseminen
 * tehdään molemmissa säikeissä ensimmäisenä.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <pthread.h>

#define VAPPUANNOSTEN_TARVE_LKM 1000

uint64_t simaa_ltr = 1000;
uint64_t munkkeja_kpl = 1000;

pthread_mutex_t mutex_sima = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mutex_munkit = PTHREAD_MUTEX_INITIALIZER;

bool juo_simaa(const char *juoja) {
    if (simaa_ltr == 0) {
        fprintf(stderr, "Ou nou! Sima on loppu!\n");
        return false;
    }
    fprintf(stdout, "%s: Yksi litra simaa juotu.\n", juoja);
    simaa_ltr--;
    return true;
}

bool ota_munkki(const char *ottaja) {
    if (munkkeja_kpl == 0) {
        fprintf(stderr, "Ou nou! Munkit on loppu!\n");
        return false;
    }
    fprintf(stdout, "%s: Yksi munkki syöty.\n", ottaja);
    munkkeja_kpl--;
    return true;
}

void * paavo(void *v) {
    const char idstr[] = "Paavo";
    int a = 0;
    for (int i = 1; i <= VAPPUANNOSTEN_TARVE_LKM; i++) {
        pthread_mutex_lock(&mutex_sima);
        pthread_mutex_lock(&mutex_munkit);
        if (juo_simaa(idstr) && ota_munkki(idstr)) {
            printf("%s: Nams, olipa hyvää. Olen saanut %d annosta.\n", idstr, ++a);
        } else {
            printf("%s: Hitsi, mulle ei riittänyt! Vappu on peruttu...\n", idstr);
        }
        pthread_mutex_unlock(&mutex_sima);
        pthread_mutex_unlock(&mutex_munkit);
    }
    printf("%s: %s Sain yhteensä %d täyttä annosta.\n",
           idstr, (a>0)?"Nams, olipa hyvää.":"", a);
    return NULL;
}

void * tomi(void *v) {
    const char idstr[] = "Tomi";
    int a = 0;
    for (int i = 1; i <= VAPPUANNOSTEN_TARVE_LKM; i++) {
        pthread_mutex_lock(&mutex_sima);
        pthread_mutex_lock(&mutex_munkit);
        if (juo_simaa(idstr) && ota_munkki(idstr)) {
            printf("%s: Nams, olipa hyvää. Olen saanut %d annosta.\n", idstr, ++a);
        } else {
            printf("%s: Hitsi, mulle ei riittänyt! Vappu on peruttu...\n", idstr);
        }
        pthread_mutex_unlock(&mutex_sima);
        pthread_mutex_unlock(&mutex_munkit);
    }
    printf("%s: %s Sain yhteensä %d täyttä annosta.\n",
           idstr, (a>0)?"Nams, olipa hyvää.":"", a);
    return NULL;
}


int main(int argc, char *argv[]) {
    int n;
    pthread_t saieA, saieB;

    if (pthread_create(&saieA, NULL, paavo, NULL) != 0) {
        fprintf(stderr, "Säikeen luonti ei onnistunut.\n");
        exit(1);
    }
    if (pthread_create(&saieB, NULL, tomi, NULL) != 0) {
        fprintf(stderr, "Säikeen luonti ei onnistunut.\n");
        exit(1);
    }
     
    if (pthread_join(saieA, NULL) != 0) {
        fprintf(stderr, "Säikeen A odottelussa tuli ongelmia.\n");
    };
    if (pthread_join(saieB, NULL) != 0) {
        fprintf(stderr, "Säikeen B odottelussa tuli ongelmia.\n");
    };

    printf("Simaa jäljellä %" PRId64 "; munkkeja jäljellä %" PRId64 "\n",
           simaa_ltr, munkkeja_kpl);
     
    return 0;
}
